/* definitions and macros for HF-Lab      1996-2015  John P. Beale */

#define VERSION_NUMBER "v0.95 16_JAN_2015 "

#ifdef __alpha
# include <float.h>
# include <math.h>
#endif

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif

#ifndef FLT_MIN
# define FLT_MIN (-9e99)
#endif

#ifndef FLT_MAX
# define FLT_MAX (9e99)
#endif

#define ADD  0           /* keytypes for image operation */
#define SUB  1
#define MUL  2
#define DIV  3
#define POW  4
#define EXP  5
#define SIN  6
#define ASIN 7
#define TAN  8
#define ATAN 9
#define LOG  10
#define INVT 11
#define ABSV 12
#define INTG 13
#define DIFF 14
#define DIF2 15
#define COM  16
#define COMN 17
#define RMAG 18
#define FLOO 19
#define CEIL 20
#define MODU 21
#define DISC 22

#define ESC '/033'
#define SEPTOR ';'     /* command separator character */
#define CLINELEN 1024   /* max number of character in a command line + 1 */
#define MAXTOK 50      /* maximum number of tokens on input line */
#define TOKLEN 80      /* maximum length of any string token */

#define AUTO 0         /* tokens for tile_mode global */
#define TOFF 1
#define TON  2

#define WIRE   0                /* 3-D shading models */
#define HEIGHT 1
#define SLOPE  2
#define LIGHT  3

#define BANDPASS 1              /* frequency-domain (fourier) filter types */
#define BANDREJECT -1
#define LOWPASS 2
#define HIGHPASS -2

#define GAUFAC 4.0              /* sigmas out along gaussian */

#define FALSE 0
#define TRUE 1

#define D double
#define U unsigned int
#define PTYPE float   /* data-type of heightfield (pixel) values */ 
#define Boolean int

#undef tolower          /* annoying detail for ANSI-compatibility */
			   
typedef struct {        /* Heightfield structure type */
   PTYPE *a;          /* 2-D array of values */
   U xsize;
   U ysize;             /* x- and y-dimensions of array */
   PTYPE min;
   PTYPE max;         /* max and min values in array */
   int c;             /* TRUE if matrix is complex, FALSE if real */
   char name[15];     /* name of HF */
 } hfield;

typedef struct Point3 {
 double x, y, z;                   /* three-dimensional point */
} PT3;

#define ABS(a)          (((a)<0) ? -(a) : (a))
#define ROUND(a)        ((a)>0 ? (int)((a)+0.5) : -(int)(0.5-(a)))
#define ZSGN(a)         (((a)<0) ? -1 : (a)>0 ? 1 : 0)  
#define SGN(a)          (((a)<0) ? -1 : 1)
#define MIN(x,y)        (((x)<(y)) ? (x) : (y) )
#define MAX(x,y)        (((x)>(y)) ? (x) : (y) )
#define ATAN2(y,x)      ((x==0 && y==0)?0:atan2(y,x))


#define INCSTACK  if (++hp == stacksize) hp = 0     /* HF stack ptr++ */
#define DECSTACK  if (--hp < 0) hp = (stacksize-1)  /* HF stack ptr-- */
#define NHF       (PTYPE *)NULL                     /* null heightfield */
#define INTERPOLATE(x,y,frac)  (1-(frac))*(x) + (frac)*(y)


/* -- system functions (?) --- */
unsigned sleep(unsigned seconds);

/* ------------------- hl.c -------------------------- */
int clear_stack(int s);            /* erase all entries in HF stack */
int pr_help(char *cmd);                   /* print help message */
void show_vars();                         /* print variable values */
void list_vars();                         /* list desc. of variables */
int set_vars(char *var, char *arg);       /* set variable values */
int keymatch(char *s,char *k, int m); /* test for keyword match */
int getaline(char s[], int lim);       /* read in one line */
int list_stack();                     /* print out contents of HF stack */
int s_empty();   /* returns true if heightfield stack is empty */
int h_newc(hfield *hfp, int xs, int ys, char *name); /* create complex HF */
int h_newr(hfield *hfp, int xs, int ys, char *name); /* create real HF */

int s_equal(char *a, char *b);  /* 1 if strings are equal */
int strnicmp_jpb(char *a, char *b, int n); /* compare n chars ignore case */
int loadhf(char *fname_in);           /* load in HF */
int savehf(char *fname_out);          /* write HF to disk */
int show_hist(int lines);
int h_tshow(int tx, int ty);   /* show matrix as text (ick) */

/* ---------- interp.c --------------------------------- */
int cmd_seq(int cmd_ptr); /* interpret string of commands starting at cmd_ptr */
int cmd_seq1(char instr[CLINELEN]);  /* run a command sequence a;b;c etc */
int h_interp(int tokens, char insp[MAXTOK][TOKLEN]); /* interpret one cmd. */
int parse_line(char *instr, int *t_ptr, char insp[MAXTOK][TOKLEN]);

/* --- dirfun.c -------------------------------------- */
void listdir(char *dname);            /* print directory entries */
void glistdir(char *dname);           /* print only graphics files */
int is_gfile(char *f);  /* returns 1 if file has known extension (eg .TGA) */
int f_delete(char *fname);            /* delete a file */

/* --- hcomp.c -------------------------------------- */
int h_minmax();                       /* set min,max values of array */
int i_minmax(hfield hf,D *mx,D *mn);  /* find min,max of imag.vals */
int r_minmax(hfield hf,D *mx,D *mn);  /* find min,max of real.vals */
int histeq(PTYPE frac);               /* histogram equalization */
int hh_hist(int bins, long **hist_p);  /* generate histogram */
int h_hist(int bins);                 /* display a histogram */
int h_hshift(PTYPE shift);            /* offset HF so hist. peak at SHIFT */
int norm(PTYPE min, PTYPE max);       /* normalize to [min..max] */
int nrm2(PTYPE max);                  /* normalize to [0..max] */
int negate();                         /* invert values in matrix */
int composit(char*o, int xo, int yo); /* HF comb. taking two operands */
int oneop(char *opn, D f1, D f2);     /* HF ops taking only one operand */
int h_slopelim(char *opn,D thresh, int iter); /* slope-dependent smoothing */
int h_diff(char *opn);                /* 1st or 2nd differential of HF */

int rotate(int deg);                  /* rotate 90,180,270 degrees */
int peak(D xpos, D ypos);             /* slew HF to locate peak at x,y */
int is_tilable(int flag);             /* true if image tiles ok */
int h_tilable(hfield h,int f);        /* 1 if HF tiles seamlessly */
int zero(int xsize, int ysize);       /* create a new blank HF */
int h_const(int xs,int ys, D val);    /* create a new constant HF */

/* ----- cplx.c functions --------------------------------- */
int c_swap();                         /* swap complex and real */
int c_join();           /* make complex from real & imag parts of matrix */
int c_split();          /* divide complex into real & imag parts */
int c_real();           /* take real part of array */
int c_diff();           /* make complex from x & y slope of matrix */
int c_int();            /* generate real mx from complex x&y diff'l */
int c_mag();            /* take magnitude of complex matrix */
int c_convert(int dir); /* convert complex Polar<->Rect representation */


/* --- hstack.c ---------------------------------------- */

int hpop();                      /* HF basic stack operations */
int hdup();
int hswap();
int hrot3();
int hrot();
int hname(char *name);          /* assign name to HF */
int hcopy(hfield *h1p,hfield h2);

/* ------------ s_graph.c  and x_graph.c --------------- */

		   /* 3D projection of heightfield (triangulation) */
int show_grid(hfield hf,           /* heightfield to display */
    char insp[MAXTOK][TOKLEN],     /* tokens array, each TOKLEN-1 long max. */
	      int tokens           /* number of tokens present */
  );

int show_flat(hfield hf, char insp[MAXTOK][TOKLEN], int tokens); /* 2D */

/* ----------------------------------------------------- */
int chdir();
void cleanup();
int startlog(char *f);          /* start log file */
int stoplog();                  /* stop log */

/* --- ops.c --------------------------------------------------- */

int time_tic();      /* store current time in time_start global */
int time_toc();      /* print difference between current time and tic_time */
int gforge1(int argc, char **argv);   /* spectral synthesis */
int gforge(int size, float h);        /* generate terrain */
int fillarray(int xsize, int ysize, float scale); /* gen 1/f noise */
int h_fft(int dir, D s);              /* complex FFT / IFFT routine */
float ran1();               /* return a single [0..1] float random # */
void seed_ran1(int seed);   /* use a single integer seed value */
int gen_rand(int xsize, int ysize);      /* generate random array */
void initgauss();       /* initialize rand #/gauss gen */
double gauss();                          /* return gaussian rand# */
double gaussn();      /* returns gaussian dist. rand # between 0 and 1 */

int h_fourfilt(D center, D Q, char *f_type);  /* filter real & imag. matrix */
void f_filter(PTYPE *real, PTYPE *imag, int xsize, int ysize, 
	      double center, double Q, int f_type);
int h_realfilt(D a1, D a2, char *f_type);  /* filter a real matrix */

int yslope(D yfrac,D scale, D slope_exp);
int h_gauss(D xfrac, D yfrac, D radfac, D hscale);  /* add gaussian hill */
int h_ring(D xfrac, D yfrac, D radfac, D width, D hscale);  /* add a ring */
int h_crater(int how_many, D ch_scale, D radius, D dfac);
int smooth(D frac);                   /* smooth heightfield */

/* ---------- ops2.c ------------------------------------------ */
int h_join(int f); /* join two HFs joined together 0=vert 1=horiz */
int h_double(D f, D f2); /* double res. with midpoint-disp interpolation */
int h_half();       /* halve resolution by averaging groups of four pixels */
int h_rhist();      /* generate radial profile histogram */
int rescale(int r1, int r2);          /* rescale heightfield */
int h_clip(int x1, int y1, int x2, int y2); /* clip out rectangular subarray */
int h_warp(int opn, D xc, D yc, D scl,int x3,int y3); /* radial map xform */
int h_cwarp(int opn, D scalefac);  /* warp with complex control */
int h_zedge(D frac, D pow);        /* set edges of HF to zero */
int h_sample(int av, char ac[MAXTOK][TOKLEN]);  /* print value of HF at x,y */
int h_print(int av, char ac[MAXTOK][TOKLEN]);  /* print all elements of HF */
int h_nsmooth(int iter, D thresh, D th2); /* elevation-dependent smoothing */

/* ---------- crater.c --------------------------------------- */

int distribute_craters(PTYPE *real, PTYPE *imag, int xsize, int ysize,
 unsigned int how_many, Boolean wrap, double ch_scale, double cr_scale,
 double dfac);

/* ------- erode.c --------------------------------- */

int h_fillb(int imax, D rate);  /* fill basin imax times */
int h_find_ua();                /* find uphill area */


/* ------ hcon.h -----------------------------------   jpb 7/15/95 */

#ifdef BIG_FLOAT
#define NUM_FMT 1
#else
#define NUM_FMT 0       /* 0 IEEE Little Endian (80x86, DEC Risc) */
#endif                  /* 1 IEEE Big Endian (SPARC, SGI, Motorola, etc.) */

#define PREC 1          /* storing Matlab file as floats rather than longs */
#define PTYPE float     /* internal representation of float numbers */
#define BYTE unsigned char
#define U unsigned int

#define Boolean int
#define FALSE 0
#define TRUE 1

#define PGM  0          /* PGM ascii type (arbitrary) */
#define TGA  1          /* POV-Ray 2.2 TGA heightfield format 16-bit */
#define PNG  2          /* PNG format (not yet supported) */
#define PG8  3          /* binary PGM format 8-bit */
#define OCT  4          /* Octave ascii format (arbitrary) */
#define MAT  5          /* Matlab Level 1.0 binary format 32bit float */
#define GIF  6          /* CompuServe (tm) GIF format 8-bit */
#define POT  7          /* Fractint, PoV-Ray "continuous potential" 16-bit */
#define RAW  8          /* RAW ascii format triangle vertices (x,y,z) */

#define MAX_PIXVAL 65536              /* ascii integer output max value */

#define El(vv, xq, yq)  vv[((yq) * xsize) + xq]  /* elmt of array */
#define El1(vv, xq, yq) vv[((yq) * xsize1)+ xq]  /* elmt of 1st array */
#define El2(vv, xq, yq) vv[((yq) * xsize2)+ xq]  /* elmt of 2nd array */
#define El3(vv, xq, yq) vv[((yq) * xsize3)+ xq]  /* elmt of 3rd array */
#define Im(vv, xq, yq)  vv[xsize*ysize + ((yq) * xsize) + xq] /* imag.part */
#define Im2(vv, xq, yq) vv[xsize2*ysize2 +((yq)*xsize2) + xq] /* imag.part */

#define Elmod(v,x,y) v[(((y)>=0)?(y)%ysize:(((y)%ysize)+ysize)%ysize)*xsize \
		     + (((x)>=0)?(x)%xsize:(((x)%xsize)+xsize)%xsize) ]
#define Elclip(v,x,y) v[(((y)<0?0:(((y)>=ysize)?(ysize-1):(y))) * xsize)\
		      + ((x)<0?0:(((x)>=xsize)?(xsize-1):(x)))]

#define X_WRAP(x)  x=(((x)>=0)?(x)%xsize:(((x)%xsize)+xsize)%xsize)
#define Y_WRAP(y)  y=(((y)>=0)?(y)%ysize:(((y)%ysize)+ysize)%ysize)

#define X_CLIP(x)  x=((x)<0?0:(((x)>=xsize)?(xsize-1):(x)))
#define Y_CLIP(y)  y=((y)<0?0:(((y)>=ysize)?(ysize-1):(y)))

			/* Matlab Level 1.0 format data file header */
typedef struct {
	long type;      /* type */
	long mrows;     /* row dimension */
	long ncols;     /* column dimension */
	long imagf;     /* flag indicating imag part */
	long namelen;    /* name length (including NULL) */
} Fmatrix;

void hf_print();                        /* print out hf array to stdout */

